@@ -1,3 +1,5 @@
+/* $XConsortium: button.c /main/75 1996/11/29 10:33:33 swick $ */
+/* $XFree86: xc/programs/xterm/button.c,v 3.11.2.1 1997/05/23 09:24:33 dawes Exp $ */
 /*
  * Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts.
  *
@@ -28,25 +30,38 @@ button.c	Handles button events in the terminal emulator.
 				J. Gettys.
 */
 
+#ifdef HAVE_CONFIG_H
+#include <xtermcfg.h>
+#endif
+
 #include "ptyx.h"		/* Xlib headers included here. */
+
 #include <X11/Xatom.h>
+
+#ifndef X_NOT_STDC_ENV
+#include <stdlib.h>
+#else
+extern char *malloc();
+#endif
+
 #include <stdio.h>
 
+#ifdef MINIX
+#include <X11/Xos.h>
+#endif
+
 #include <X11/Xmu/Atoms.h>
 #include <X11/Xmu/StdSel.h>
 
 #include "data.h"
 #include "error.h"
 #include "menu.h"
+
+#include "xterm.h"
+
+#ifdef HANGUL
 #include "hangul.h"
-
-void TrackText();
-
-extern char *malloc();
-
-extern void input_to_chat();
-extern void HideCursor();
-extern void DoSecureKeyboard();
+#endif
 
 #define KeyState(x) (((x) & (ShiftMask|ControlMask)) + (((x) & Mod1Mask) ? 2 : 0))
     /* adds together the bits:
@@ -64,17 +79,6 @@ extern void DoSecureKeyboard();
 
 extern char *xterm_name;
 
-static void PointToRowCol();
-static void SelectionReceived();
-static void TrackDown();
-static void ComputeSelect();
-static void EditorButton();
-static void ExtendExtend();
-static void ReHiliteText();
-static void SelectSet();
-static void StartSelect();
-static int Length();
-static char *SaveText();
 
 extern XtermWidget term;
 
@@ -100,7 +104,7 @@ static int saveStartRRow, saveStartRCol, saveEndRRow, saveEndRCol;
 
 /* Multi-click handling */
 static int numberOfClicks = 0;
-static long int lastButtonUpTime = 0;
+static Time lastButtonUpTime = 0;
 typedef int SelectUnit;
 #define SELECTCHAR 0
 #define SELECTWORD 1
@@ -111,6 +115,33 @@ static SelectUnit selectUnit;
 /* Send emacs escape code when done selecting or extending? */
 static int replyToEmacs;
 
+static Boolean ConvertSelection PROTO_XT_CVT_SELECT_ARGS;
+static SelectUnit EvalSelectUnit PROTO((Time buttonDownTime, SelectUnit defaultUnit));
+static char *SaveText PROTO((TScreen *screen, int row, int scol, int ecol, char *lp, int *eol));
+static int LastTextCol PROTO((int row));
+static int Length PROTO((TScreen *screen, int row, int scol, int ecol));
+static void ComputeSelect PROTO((int startRow, int startCol, int endRow, int endCol, Bool extend));
+static void EditorButton PROTO((XButtonEvent *event));
+static void EndExtend PROTO((Widget w, XEvent *event, String *params, Cardinal num_params, Bool use_cursor_loc));
+static void ExtendExtend PROTO((int row, int col));
+static void LoseSelection PROTO((Widget w, Atom *selection));
+static void PointToRowCol PROTO((int y, int x, int *r, int *c));
+static void ReHiliteText PROTO((int frow, int fcol, int trow, int tcol));
+static void SaltTextAway PROTO((int crow, int ccol, int row, int col, String *params, Cardinal num_params));
+static void SelectSet PROTO((Widget w, XEvent *event, String *params, Cardinal num_params));
+static void SelectionDone PROTO((Widget w, Atom *selection, Atom *target));
+static void SelectionReceived PROTO_XT_SEL_CB_ARGS;
+static void StartSelect PROTO((int startrow, int startcol));
+static void TrackDown PROTO((XButtonEvent *event));
+static void _GetSelection PROTO((Widget w, Time ev_time, String *params, Cardinal num_params));
+static void _OwnSelection PROTO((XtermWidget termw, String *selections, Cardinal count));
+static void do_select_end PROTO((Widget w, XEvent *event, String *params, Cardinal *num_params, Bool use_cursor_loc));
+static void do_select_start PROTO((Widget w, XEvent *event, int startrow, int startcol));
+static void do_start_extend PROTO((Widget w, XEvent *event, String *params, Cardinal *num_params, Bool use_cursor_loc));
+
+#ifdef HANGUL
+static int conv_ksc_to_ct PROTO((Char *str, int len, Char *des));
+#endif
 
 Boolean SendMousePosition(w, event)
 Widget w;
@@ -136,7 +167,7 @@ XEvent* event;
 
 	if (KeyModifiers == 0) {
 	    if (event->type == ButtonPress)
-		EditorButton(event);
+		EditorButton((XButtonEvent *)event);
 	    return True;
 	}
 	return False;
@@ -144,7 +175,7 @@ XEvent* event;
       case 2: /* DEC vt200 compatible */
 
 	if (KeyModifiers == 0 || KeyModifiers == ControlMask) {
-	    EditorButton(event);
+	    EditorButton((XButtonEvent *)event);
 	    return True;
 	}
 	return False;
@@ -153,11 +184,11 @@ XEvent* event;
 	if (  event->type == ButtonPress &&
 	      KeyModifiers == 0 &&
 	      event->xbutton.button == Button1 ) {
-	    TrackDown(event);
+	    TrackDown((XButtonEvent *)event);
 	    return True;
 	}
 	if (KeyModifiers == 0 || KeyModifiers == ControlMask) {
-	    EditorButton(event);
+	    EditorButton((XButtonEvent *)event);
 	    return True;
 	}
 	/* fall through */
@@ -168,6 +199,62 @@ XEvent* event;
 #undef KeyModifiers
 }
 
+void
+DiredButton(w, event, params, num_params)
+Widget w;
+XEvent *event;			/* must be XButtonEvent */
+String *params;			/* selections */
+Cardinal *num_params;
+{	/* ^XM-G<line+' '><col+' '> */
+	register TScreen *screen = &term->screen;
+	int pty = screen->respond;
+	char Line[ 6 ];
+	register unsigned line, col;
+
+    if (event->type != ButtonPress && event->type != ButtonRelease)
+    	return;
+    strcpy( Line, "\030\033G  " );
+
+	line = ( event->xbutton.y - screen->border ) / FontHeight( screen );
+	col = (event->xbutton.x - screen->border - Scrollbar(screen))
+	 / FontWidth(screen);
+	Line[3] = ' ' + col;
+	Line[4] = ' ' + line;
+	v_write(pty, Line, 5 );
+}
+
+void
+ViButton(w, event, params, num_params)
+Widget w;
+XEvent *event;			/* must be XButtonEvent */
+String *params;			/* selections */
+Cardinal *num_params;
+{	/* ^XM-G<line+' '><col+' '> */
+	register TScreen *screen = &term->screen;
+	int pty = screen->respond;
+	char Line[ 6 ];
+	register int line;
+
+    if (event->type != ButtonPress && event->type != ButtonRelease)
+    	return;
+
+	line = screen->cur_row -
+		(( event->xbutton.y - screen->border ) / FontHeight( screen ));
+/* fprintf( stderr, "xtdb line=%d\n", line ); */
+	if ( ! line ) return;
+	Line[ 1 ] = 0;
+	Line[ 0 ] = 27;
+	v_write(pty, Line, 1 );
+
+	Line[ 0 ] = 'p' & 0x1f;
+
+	if ( line < 0 )
+	{	line = -line;
+		Line[ 0 ] = 'n' & 0x1f;
+	}
+	while ( --line >= 0 ) v_write(pty, Line, 1 );
+}
+
 
 /*ARGSUSED*/
 void HandleSelectExtend(w, event, params, num_params)
@@ -193,8 +280,6 @@ Cardinal *num_params;		/* unused */
 	}
 }
 
-static void EndExtend();
-
 static void do_select_end (w, event, params, num_params, use_cursor_loc)
 Widget w;
 XEvent *event;			/* must be XButtonEvent */
@@ -244,9 +329,9 @@ struct _SelectionList {
 };
 
 
-static void _GetSelection(w, time, params, num_params)
+static void _GetSelection(w, ev_time, params, num_params)
 Widget w;
-Time time;
+Time ev_time;
 String *params;			/* selections in precedence order */
 Cardinal num_params;
 {
@@ -277,17 +362,17 @@ Cardinal num_params;
 	    SelectionReceived(w, NULL, &selection, &type, (XtPointer)line,
 			      &nbytes, &fmt8);
 	else if (num_params > 1)
-	    _GetSelection(w, time, params+1, num_params-1);
+	    _GetSelection(w, ev_time, params+1, num_params-1);
     } else {
 	struct _SelectionList* list;
 	if (--num_params) {
 	    list = XtNew(struct _SelectionList);
 	    list->params = params + 1;
 	    list->count = num_params; /* decremented above */
-	    list->time = time;
+	    list->time = ev_time;
 	} else list = NULL;
 	XtGetSelectionValue(w, selection, XA_STRING, SelectionReceived,
-			    (XtPointer)list, time);
+			    (XtPointer)list, ev_time);
     }
 }
 
@@ -304,9 +389,12 @@ unsigned long *length;
 int *format;
 {
     int pty = ((XtermWidget)w)->screen.respond;	/* file descriptor of pty */
+    TScreen *screen = &(((XtermWidget)w)->screen);
     register char *lag, *cp, *end;
     char *line = (char*)value;
+#ifdef HANGUL
     int first;
+#endif
 				  
     if (*type == 0 /*XT_CONVERT_FAIL*/ || *length == 0 || value == NULL) {
 	/* could not get this selection, so see if there are more to try */
@@ -324,31 +412,41 @@ int *format;
 
     end = &line[*length];
     lag = line;
+#ifdef HANGUL
     first = 1;
+    if(screen->code == C_JOHAB) {
+	convert_ks_to_johab(line, line, end - line + 1);
+    }
+#endif
     for (cp = line; cp != end; cp++)
 	{
 	    if (*cp != '\n') continue;
 	    *cp = '\r';
+#ifdef HANGUL
 	    if (first) {
 		if (chat_mode) {
 		    chat_flush(0);
 		}
 		first = 0;
 	    }
+#endif
 	    v_write(pty, lag, cp - lag + 1);
 	    lag = cp + 1;
 	}
+#ifdef HANGUL
     if (lag != end) {
 	if (chat_mode) {
-	    char buf[1024];
-	    int len;
-
-	    len = convert_ks_to_johab(lag, buf, end - lag);
-	    input_to_chat(buf, len);
-	}
-	else
+	    /*char buf[1024];
+	    int len = convert_ks_to_johab(lag, buf, end - lag);
+	    */
+	    input_to_chat(lag, end - lag);
+	} else
 	    v_write(pty, lag, end - lag);
     }
+#else
+    if (lag != end)
+    	v_write(pty, lag, end - lag);
+#endif
 
     XtFree(client_data);
     XtFree(value);
@@ -367,20 +465,27 @@ Cardinal *num_params;
 }
 
 
-static void
-SetSelectUnit(buttonDownTime, defaultUnit)
+static SelectUnit
+EvalSelectUnit(buttonDownTime, defaultUnit)
     Time buttonDownTime;
     SelectUnit defaultUnit;
 {
-/* Do arithmetic as integers, but compare as unsigned solves clock wraparound */
-	if ((long unsigned)((long int)buttonDownTime - lastButtonUpTime)
-	 > term->screen.multiClickTime) {
-		numberOfClicks = 1;
-		selectUnit = defaultUnit;
-	} else {
-		++numberOfClicks;
-		selectUnit = ((selectUnit + 1) % NSELECTUNITS);
-	}
+    int delta;
+
+    if (lastButtonUpTime == (Time) 0) /* first time and once in a blue moon */
+	delta = term->screen.multiClickTime + 1;
+    else if (buttonDownTime > lastButtonUpTime) /* most of the time */
+	delta = buttonDownTime - lastButtonUpTime;
+    else /* time has rolled over since lastButtonUpTime */
+	delta = (((Time) ~0) - lastButtonUpTime) + buttonDownTime;
+
+    if (delta > term->screen.multiClickTime) {
+	numberOfClicks = 1;
+	return defaultUnit;
+    } else {
+	++numberOfClicks;
+	return ((selectUnit + 1) % NSELECTUNITS);
+    }
 }
 
 static void do_select_start (w, event, startrow, startcol)
@@ -389,7 +494,7 @@ XEvent *event;			/* must be XButtonEvent* */
 int startrow, startcol;
 {
 	if (SendMousePosition(w, event)) return;
-	SetSelectUnit(event->xbutton.time, SELECTCHAR);
+	selectUnit = EvalSelectUnit(event->xbutton.time, SELECTCHAR);
 	replyToEmacs = FALSE;
 	StartSelect(startrow, startcol);
 }
@@ -403,34 +508,22 @@ String *params;			/* unused */
 Cardinal *num_params;		/* unused */
 {
 	register TScreen *screen = &((XtermWidget)w)->screen;
-	int startrow, startcol, row;
-        int xOffset_mode = 0;
-        int xOffset_code = 0;
-        int xOffset_layout = 0;
+	int startrow, startcol;
 
 	firstValidRow = 0;
 	lastValidRow  = screen->max_row;
 	PointToRowCol(event->xbutton.y, event->xbutton.x, &startrow, &startcol);
-	row = (event->xbutton.y - screen->border) / FontHeight(screen);
-	if (row == lastValidRow + 1) {
-            xOffset_mode = ST_MODE_X;
-            xOffset_code = ST_MODE_X + ST_MODE_LEN;
-            xOffset_layout = xOffset_code + ((screen->showStatus.hideCode) ? 0 : ST_CODE_LEN);
-
-	    if (startcol >= xOffset_mode &&
-		startcol < (xOffset_mode + ST_MODE_LEN))
-		HandleToggleHangul();
-	    else if (startcol >= (xOffset_code) &&
-		     startcol < (xOffset_code + ST_CODE_LEN) &&
-                     !(screen->showStatus.hideCode))
-		HandleChangeCode();
-	    else if (startcol >= xOffset_layout &&
-		     startcol < xOffset_layout + ST_KBD_LEN &&
-                     !(screen->showStatus.hideHanLayout))
-		HandleChangeKeyboard();
-	}
-	else
-	    do_select_start (w, event, startrow, startcol);
+#ifdef HANGUL
+	if ( (event->xbutton.y - screen->border)/FontHeight(screen) >= (lastValidRow + 1)) {
+	    if (is_mode_button(startcol))
+		HandleToggleHangul(w, event, params, num_params);
+	    else if (is_code_button(startcol))
+		HandleChangeCode(w, event, params, num_params);
+	    else if (is_kbd_button(startcol))
+		HandleChangeKeyboard(w, event, params, num_params);
+	} else
+#endif
+	do_select_start (w, event, startrow, startcol);
 }
 
 
@@ -454,14 +547,14 @@ TrackDown(event)
 {
 	int startrow, startcol;
 
-	SetSelectUnit(event->time, SELECTCHAR);
+	selectUnit = EvalSelectUnit(event->time, SELECTCHAR);
 	if (numberOfClicks > 1 ) {
 		PointToRowCol(event->y, event->x, &startrow, &startcol);
 		replyToEmacs = TRUE;
 		StartSelect(startrow, startcol);
 	} else {
 		waitingForTrackInfo = 1;
-		EditorButton(event);
+		EditorButton((XButtonEvent *)event);
 	}
 }
 
@@ -471,8 +564,9 @@ TrackDown(event)
 			else if (x >= screen->max_row) \
 			    x = screen->max_row;
 
-void TrackMouse(func, startrow, startcol, firstrow, lastrow)
-int func, startrow, startcol, firstrow, lastrow;
+void
+TrackMouse(func, startrow, startcol, firstrow, lastrow)
+    int func, startrow, startcol, firstrow, lastrow;
 {
 	TScreen *screen = &term->screen;
 
@@ -530,9 +624,9 @@ EndExtend(w, event, params, num_params, use_cursor_loc)
     Cardinal num_params;
     Bool use_cursor_loc;
 {
-	int	row, col;
+	int	row, col, count;
 	TScreen *screen = &term->screen;
-	char line[9];
+	unsigned char line[9];
 
 	if (use_cursor_loc) {
 	    row = screen->cursor_row;
@@ -544,24 +638,30 @@ EndExtend(w, event, params, num_params, use_cursor_loc)
 	lastButtonUpTime = event->xbutton.time;
 	if (startSRow != endSRow || startSCol != endSCol) {
 		if (replyToEmacs) {
+			count = 0;
+			if (screen->control_eight_bits) {
+				line[count++] = CSI;
+			} else {
+				line[count++] = ESC;
+				line[count++] = '[';
+			}
 			if (rawRow == startSRow && rawCol == startSCol 
 			    && row == endSRow && col == endSCol) {
 			 	/* Use short-form emacs select */
-				strcpy(line, "\033[t");
-				line[3] = ' ' + endSCol + 1;
-				line[4] = ' ' + endSRow + 1;
-				v_write(screen->respond, line, 5);
+				line[count++] = 't';
+				line[count++] = ' ' + endSCol + 1;
+				line[count++] = ' ' + endSRow + 1;
 			} else {
 				/* long-form, specify everything */
-				strcpy(line, "\033[T");
-				line[3] = ' ' + startSCol + 1;
-				line[4] = ' ' + startSRow + 1;
-				line[5] = ' ' + endSCol + 1;
-				line[6] = ' ' + endSRow + 1;
-				line[7] = ' ' + col + 1;
-				line[8] = ' ' + row + 1;
-				v_write(screen->respond, line, 9);
+				line[count++] = 'T';
+				line[count++] = ' ' + startSCol + 1;
+				line[count++] = ' ' + startSRow + 1;
+				line[count++] = ' ' + endSCol + 1;
+				line[count++] = ' ' + endSRow + 1;
+				line[count++] = ' ' + col + 1;
+				line[count++] = ' ' + row + 1;
 			}
+			v_write(screen->respond, (char *)line, count);
 			TrackText(0, 0, 0, 0);
 		}
 	}
@@ -579,8 +679,6 @@ HandleSelectSet(w, event, params, num_params)
 	SelectSet (w, event, params, *num_params);
 }
 
-static void SaltTextAway();
-
 /* ARGSUSED */
 static void
 SelectSet (w, event, params, num_params)
@@ -613,7 +711,7 @@ Bool use_cursor_loc;
 	if (SendMousePosition(w, event)) return;
 	firstValidRow = 0;
 	lastValidRow  = screen->max_row;
-	SetSelectUnit(event->xbutton.time, selectUnit);
+	selectUnit = EvalSelectUnit(event->xbutton.time, selectUnit);
 	replyToEmacs = FALSE;
 
 	if (numberOfClicks == 1) {
@@ -703,10 +801,7 @@ Cardinal *num_params;		/* unused */
     do_start_extend (w, event, params, num_params, True);
 }
 
-
-
-
-
+void
 ScrollSelection(screen, amount)
 register TScreen* screen;
 register int amount;
@@ -716,15 +811,15 @@ register int amount;
     register int maxcol = screen->max_col;
 
 #define scroll_update_one(row, col) \
-        row += amount; \
-        if (row < minrow) { \
-            row = minrow; \
-            col = 0; \
-        } \
-        if (row > maxrow) { \
-            row = maxrow; \
-            col = maxcol; \
-        }
+    	row += amount; \
+	if (row < minrow) { \
+	    row = minrow; \
+	    col = 0; \
+	} \
+	if (row > maxrow) { \
+	    row = maxrow; \
+	    col = maxcol; \
+	}
 
     scroll_update_one(startRRow, startRCol);
     scroll_update_one(endRRow, endRCol);
@@ -736,57 +831,13 @@ register int amount;
     scroll_update_one(screen->startHRow, screen->startHCol);
     scroll_update_one(screen->endHRow, screen->endHCol);
 
-#ifdef DELETED
-    register int minrow = -screen->savedlines;
-
-    /* Sent by scrollbar stuff, so amount never takes selection out of
-       saved text */
-
-    /* XXX - the preceeding is false; cat /etc/termcap (or anything
-       larger than the number of saved lines plus the screen height) and then
-       hit extend select */
-
-    startRRow += amount; endRRow += amount;
-    startSRow += amount; endSRow += amount;
-    rawRow += amount;
-    screen->startHRow += amount;
-    screen->endHRow += amount;
-
-    if (startRRow < minrow) {
-	startRRow = minrow;
-	startRCol = 0;
-    }
-    if (endRRow < minrow) {
-	endRRow = minrow;
-        endRCol = 0;
-    }
-    if (startSRow < minrow) {
-	startSRow = minrow;
-	startSCol = 0;
-    }
-    if (endSRow < minrow) {
-	endSRow = minrow;
-	endSCol = 0;
-    }
-    if (rawRow < minrow) {
-	rawRow = minrow;
-	rawCol = 0;
-    }
-    if (screen->startHRow < minrow) {
-	screen->startHRow = minrow;
-	screen->startHCol = 0;
-    }
-    if (screen->endHRow < minrow) {
-	screen->endHRow = minrow;
-	screen->endHCol = 0;
-    }
-#endif
     screen->startHCoord = Coordinate (screen->startHRow, screen->startHCol);
     screen->endHCoord = Coordinate (screen->endHRow, screen->endHCol);
 }
 
 
 /*ARGSUSED*/
+void
 ResizeSelection (screen, rows, cols)
     TScreen *screen;
     int rows, cols;
@@ -817,29 +868,23 @@ PointToRowCol(y, x, r, c)
        maximum value. */
 {
 	register TScreen *screen = &term->screen;
-	register row, col, i;
-	register Char *p;
+	register row, col;
 
 	row = (y - screen->border) / FontHeight(screen);
 	if(row < firstValidRow)
 		row = firstValidRow;
 	else if(row > lastValidRow)
 		row = lastValidRow;
-	col = (x - screen->border - screen->scrollbar) / FontWidth(screen);
+    col = (x - OriginX(screen)) / FontWidth(screen);
 	if(col < 0)
 		col = 0;
 	else if(col > screen->max_col+1) {
 		col = screen->max_col+1;
 	}
-
-#if !defined(NO_PATCH_ADJUSTHANGULSTRING) /* by hanmaum 1998.2.4 */
-	col -= adjust_hangul_string(screen->buf[4 * (row + screen->topline)],
-				    col, C_JOHAB);
-#else
-	col -= adjust_hangul_string(screen->buf[4 * (row + screen->topline)],
+#ifdef HANGUL
+	col -= adjust_hangul_string(SCRN_BUF_CHARS(screen,row + screen->topline),
 				    col);
 #endif
-
 	*r = row;
 	*c = col;
 }
@@ -853,7 +898,7 @@ LastTextCol(row)
 	register Char *ch;
 
 	for ( i = screen->max_col,
-	        ch = screen->buf[4 * (row + screen->topline) + 1] + i ;
+	        ch = SCRN_BUF_ATTRS(screen, (row + screen->topline)) + i ;
 	      i >= 0 && !(*ch & CHARDRAWN) ;
 	      ch--, i--)
 	    ;
@@ -870,6 +915,7 @@ LastTextCol(row)
 **	- binding characters	[0x40,0x7f] U [0xc0,0xff]
 **  	- execeptions
 */
+#ifndef HANGUL
 static int charClass[256] = {
 /* NUL  SOH  STX  ETX  EOT  ENQ  ACK  BEL */
     32,   1,   1,   1,   1,   1,   1,   1,
@@ -935,6 +981,73 @@ static int charClass[256] = {
     48,  48,  48,  48,  48,  48,  48,  248,
 /*  o/   u`   u'   u^   u:   y'    P   y: */
     48,  48,  48,  48,  48,  48,  48,  48};
+#else
+static int charClass[256] = {
+/* NUL  SOH  STX  ETX  EOT  ENQ  ACK  BEL */
+    32,   1,   1,   1,   1,   1,   1,   1,
+/*  BS   HT   NL   VT   NP   CR   SO   SI */
+     1,  32,   1,   1,   1,   1,   1,   1,
+/* DLE  DC1  DC2  DC3  DC4  NAK  SYN  ETB */
+     1,   1,   1,   1,   1,   1,   1,   1,
+/* CAN   EM  SUB  ESC   FS   GS   RS   US */
+     1,   1,   1,   1,   1,   1,   1,   1,
+/*  SP    !    "    #    $    %    &    ' */
+    32,  33,  34,  35,  36,  37,  38,  39,
+/*   (    )    *    +    ,    -    .    / */
+    40,  41,  42,  43,  44,  45,  46,  47,
+/*   0    1    2    3    4    5    6    7 */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   8    9    :    ;    <    =    >    ? */
+    48,  48,  58,  59,  60,  61,  62,  63,
+/*   @    A    B    C    D    E    F    G */
+    64,  48,  48,  48,  48,  48,  48,  48,
+/*   H    I    J    K    L    M    N    O */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   P    Q    R    S    T    U    V    W */ 
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   X    Y    Z    [    \    ]    ^    _ */
+    48,  48,  48,  91,  92,  93,  94,  48,
+/*   `    a    b    c    d    e    f    g */
+    96,  48,  48,  48,  48,  48,  48,  48,
+/*   h    i    j    k    l    m    n    o */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   p    q    r    s    t    u    v    w */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   x    y    z    {    |    }    ~  DEL */
+    48,  48,  48, 123, 124, 125, 126,   1,
+/* x80  x81  x82  x83  IND  NEL  SSA  ESA */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/* HTS  HTJ  VTS  PLD  PLU   RI  SS2  SS3 */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/* DCS  PU1  PU2  STS  CCH   MW  SPA  EPA */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/* x98  x99  x9A  CSI   ST  OSC   PM  APC */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   -    i   c/    L   ox   Y-    |   So */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  ..   c0   ip   <<    _        R0    - */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   o   +-    2    3    '    u   q|    . */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   ,    1    2   >>  1/4  1/2  3/4    ? */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  A`   A'   A^   A~   A:   Ao   AE   C, */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  E`   E'   E^   E:   I`   I'   I^   I: */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  D-   N~   O`   O'   O^   O~   O:    X */ 
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  O/   U`   U'   U^   U:   Y'    P    B */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  a`   a'   a^   a~   a:   ao   ae   c, */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  e`   e'   e^   e:    i`  i'   i^   i: */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*   d   n~   o`   o'   o^   o~   o:   -: */
+    48,  48,  48,  48,  48,  48,  48,  48,
+/*  o/   u`   u'   u^   u:   y'    P   y: */
+    48,  48,  48,  48,  48,  48,  48,  48};
+#endif
 
 int SetCharacterClassRange (low, high, value)
     register int low, high;		/* in range of [0..255] */
@@ -952,13 +1065,14 @@ int SetCharacterClassRange (low, high, value)
  * sets startSRow startSCol endSRow endSCol
  * ensuring that they have legal values
  */
+
 static void
 ComputeSelect(startRow, startCol, endRow, endCol, extend)
     int startRow, startCol, endRow, endCol;
     Bool extend;
 {
 	register TScreen *screen = &term->screen;
-	register Char *ptr, c;
+	register Char *ptr;
 	register int length;
 	register int class;
 	int osc = startSCol;
@@ -991,33 +1105,54 @@ ComputeSelect(startRow, startCol, endRow, endCol, extend)
 				startSCol = 0;
 				startSRow++;
 			} else {
-				ptr = screen->buf[4*(startSRow+screen->topline)]
+				ptr = SCRN_BUF_CHARS(screen, startSRow+screen->topline)
 				 + startSCol;
 				class = charClass[*ptr];
-				c = *ptr;
+#ifndef HANGUL
 				do {
 					--startSCol;
 					--ptr;
 				} while (startSCol >= 0
-				 && (charClass[*ptr] == class ||
-				     *ptr & 0x80 && c & 0x80));
-				++startSCol;
+				 && charClass[*ptr] == class);
+#else
+				{
+				char *attr = 
+					SCRN_BUF_ATTRS(screen,startSRow+screen->topline) + startSCol;
+				length=LastTextCol(startSRow);
+				attr--;
+				do {
+					ptr--;
+					attr--;
+					startSCol--;
+					if(startSCol > 0 && *(attr) & FIRST_HAN && 
+					   charClass[*(ptr-1)] == class) {
+						ptr--;startSCol--;
+						attr--;
+					}
+				} while(charClass[*ptr] == class && startSCol >= 0);
+				}
+#endif
+				startSCol++;
 			}
 			if (endSCol > (LastTextCol(endSRow) + 1)) {
 				endSCol = 0;
 				endSRow++;
 			} else {
 				length = LastTextCol(endSRow);
-				ptr = screen->buf[4*(endSRow+screen->topline)]
+				ptr = SCRN_BUF_CHARS(screen, endSRow+screen->topline)
 				 + endSCol;
 				class = charClass[*ptr];
-				c = *ptr;
 				do {
+#ifdef HANGUL
+
+					if(*ptr & 0x80 && charClass[*ptr] == class) {
+						ptr++;endSCol++;
+					}
+#endif
 					++endSCol;
 					++ptr;
 				} while (endSCol <= length
-				 && (charClass[*ptr] == class ||
-				     *ptr & 0x80 && c & 0x80));
+				 && charClass[*ptr] == class);
 				/* Word select selects if pointing to any char
 				   in "word", especially in that it includes
 				   the last character in a word.  So no --endSCol
@@ -1047,7 +1182,8 @@ ComputeSelect(startRow, startCol, endRow, endCol, extend)
 	return;
 }
 
-void TrackText(frow, fcol, trow, tcol)
+void
+TrackText(frow, fcol, trow, tcol)
     register int frow, fcol, trow, tcol;
     /* Guaranteed (frow, fcol) <= (trow, tcol) */
 {
@@ -1098,6 +1234,7 @@ ReHiliteText(frow, fcol, trow, tcol)
 {
 	register TScreen *screen = &term->screen;
 	register int i;
+		
 
 	if (frow < 0)
 	    frow = fcol = 0;
@@ -1128,8 +1265,6 @@ ReHiliteText(frow, fcol, trow, tcol)
 	}
 }
 
-static void _OwnSelection();
-
 static void
 SaltTextAway(crow, ccol, row, col, params, num_params)
     /*register*/ int crow, ccol, row, col;
@@ -1164,9 +1299,13 @@ SaltTextAway(crow, ccol, row, col, params, num_params)
 	/* now get some memory to save it in */
 
 	if (screen->selection_size <= j) {
+#ifndef HANGUL
+	    if((line = malloc((unsigned) j + 1)) == (char *)NULL)
+#else
 	    if((line = malloc(((unsigned) j) * 4 + 1)) == (char *)NULL)
 				/* ϱ 4 ϴ  ϼ  
 				    ȯɶ ִ 4 þ  */
+#endif
 		SysError(ERROR_BMALLOC2);
 	    XtFree(screen->selection);
 	    screen->selection = line;
@@ -1217,15 +1356,14 @@ int *format;
 		    target, type, (caddr_t*)&std_targets, &std_length, format
 		   );
 	*length = std_length + 5;
-	*value = (XtPointer)XtMalloc(sizeof(Atom)*(*length));
-	targetP = *(Atom**)value;
+	targetP = (Atom*)XtMalloc(sizeof(Atom)*(*length));
+	*value = (XtPointer) targetP;
 	*targetP++ = XA_STRING;
 	*targetP++ = XA_TEXT(d);
 	*targetP++ = XA_COMPOUND_TEXT(d);
 	*targetP++ = XA_LENGTH(d);
 	*targetP++ = XA_LIST_LENGTH(d);
-	/* no overlap - dkim@surecom.com */
-	memcpy((char*)targetP, (char*)std_targets, sizeof(Atom)*std_length);
+	memcpy ( (char*)targetP, (char*)std_targets, sizeof(Atom)*std_length);
 	XtFree((char*)std_targets);
 	*type = XA_ATOM;
 	*format = 32;
@@ -1233,25 +1371,50 @@ int *format;
     }
 
     if (*target == XA_STRING ||
+#ifndef HANGUL
 	*target == XA_TEXT(d) ||
 	*target == XA_COMPOUND_TEXT(d)) {
 	if (*target == XA_COMPOUND_TEXT(d))
 	    *type = *target;
 	else
 	    *type = XA_STRING;
+#else
+	*target == XA_TEXT(d)) {
+	*type = XA_STRING;
+#endif
 	*value = xterm->screen.selection;
 	*length = xterm->screen.selection_length;
 	*format = 8;
 	return True;
     }
+#ifdef HANGUL
+    if (*target == XA_COMPOUND_TEXT(d)) {
+	    static int ctbuf_len = 0;
+	    static char *ctbuf;
+		int ctlen;
+		int len;
+	    len = xterm->screen.selection_length;
+		ctlen = conv_ksc_to_ct((Char *)xterm->screen.selection, len, (Char *)NULL);
+	    if(ctbuf_len < ctlen) {
+		    if (ctbuf_len) XtFree(ctbuf);
+		    ctbuf = XtMalloc(ctlen);
+		    ctbuf_len = ctlen;
+	    }
+		conv_ksc_to_ct((Char *)xterm->screen.selection, len, (Char *)ctbuf);
+	    *type = *target;
+	    *value = ctbuf;
+	    *length = ctlen;
+	    *format = 8;
+	    return True;
+    }
+#endif
     if (*target == XA_LIST_LENGTH(d)) {
 	*value = XtMalloc(4);
 	if (sizeof(long) == 4)
 	    *(long*)*value = 1;
 	else {
 	    long temp = 1;
-	    /* no overlap - dkim@surecom.com */
-	    memcpy((char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
+	    memcpy ( (char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
 	}
 	*type = XA_INTEGER;
 	*length = 1;
@@ -1264,8 +1427,7 @@ int *format;
 	    *(long*)*value = xterm->screen.selection_length;
 	else {
 	    long temp = xterm->screen.selection_length;
-	    /* no overlap - dkim@surecom.com */
-	    memcpy((char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
+	    memcpy ( (char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
 	}
 	*type = XA_INTEGER;
 	*length = 1;
@@ -1289,7 +1451,7 @@ static void LoseSelection(w, selection)
 {
     register TScreen* screen = &((XtermWidget)w)->screen;
     register Atom* atomP;
-    int i;
+    Cardinal i;
     for (i = 0, atomP = screen->selection_atoms;
 	 i < screen->selection_count; i++, atomP++)
     {
@@ -1339,7 +1501,7 @@ static void _OwnSelection(termw, selections, count)
     Cardinal count;
 {
     Atom* atoms = termw->screen.selection_atoms;
-    int i;
+    Cardinal i;
     Boolean have_selection = False;
 
     if (termw->screen.selection_length < 0) return;
@@ -1386,13 +1548,13 @@ static void _OwnSelection(termw, selections, count)
 	TrackText(0, 0, 0, 0);
 }
 
-/* void */
+void
 DisownSelection(termw)
     register XtermWidget termw;
 {
     Atom* atoms = termw->screen.selection_atoms;
     Cardinal count = termw->screen.selection_count;
-    int i;
+    Cardinal i;
 
     for (i = 0; i < count; i++) {
 	int cutbuffer;
@@ -1441,7 +1603,7 @@ SaveText(screen, row, scol, ecol, lp, eol)
     int *eol;
 {
 	register int i = 0;
-	register Char *ch = screen->buf[4 * (row + screen->topline)];
+	register Char *ch = SCRN_BUF_CHARS(screen, row + screen->topline);
 	Char attr;
 	register int c;
 
@@ -1468,16 +1630,19 @@ SaveText(screen, row, scol, ecol, lp, eol)
 				c = '#'; /* char on screen is pound sterling */
 			else
 				c += 0x5f; /* char is from DEC drawing set */
-		} else if(c == 0x7f) {
+		} else if(c == 0x7f)
 			c = 0x5f;
-		}
-		if (c & 0x80) {
+#ifdef HANGUL
+		if ((c & 0x80) && screen->code == C_JOHAB) {
 		    lp += convert_johab_to_ks(&ch[i], lp, 2);
 				/*  ϼθ ȯ */
 		    i++;
 		} else {
 		    *lp++ = c;
 		}
+#else
+		*lp++ = c;
+#endif
 	}
 	return(lp);
 }
@@ -1488,28 +1653,62 @@ EditorButton(event)
 {
 	register TScreen *screen = &term->screen;
 	int pty = screen->respond;
-	char line[6];
+	char unsigned line[6];
 	register unsigned row, col;
-	int button; 
+	int button, count = 0;
 
 	button = event->button - 1; 
 
 	row = (event->y - screen->border) 
 	 / FontHeight(screen);
-	col = (event->x - screen->border - screen->scrollbar)
+    col = (event->x - OriginX(screen))
 	 / FontWidth(screen);
-	(void) strcpy(line, "\033[M");
-	if (screen->send_mouse_pos == 1) {
-		line[3] = ' ' + button;
+	if (screen->control_eight_bits) {
+		line[count++] = CSI;
 	} else {
-		line[3] = ' ' + (KeyState(event->state) << 2) + 
+		line[count++] = ESC;
+		line[count++] = '[';
+	}
+	line[count++] = 'M';
+	if (screen->send_mouse_pos == 1) {
+		line[count++] = ' ' + button;
+	} else {
+		line[count++] = ' ' + (KeyState(event->state) << 2) + 
 			((event->type == ButtonPress)? button:3);
 	}
-	line[4] = ' ' + col + 1;
-	line[5] = ' ' + row + 1;
-	v_write(pty, line, 6);
+	line[count++] = ' ' + col + 1;
+	line[count++] = ' ' + row + 1;
+	v_write(pty, (char *)line, count);
 }
 
+
+#ifdef TEK_MODE
+/*ARGSUSED*/
+void HandleGINInput (w, event, param_list, nparamsp)
+    Widget w;
+    XEvent *event;
+    String *param_list;
+    Cardinal *nparamsp;
+{
+    if (term->screen.TekGIN && *nparamsp == 1) {
+	int c = param_list[0][0];
+	switch (c) {
+	  case 'l': case 'm': case 'r':
+	  case 'L': case 'M': case 'R':
+	    break;
+	  default:
+	    Bell (XkbBI_MinorError,0);	/* let them know they goofed */
+	    c = 'l';				/* provide a default */
+	}
+	TekEnqMouse (c | 0x80);
+	TekGINoff();
+    } else {
+	Bell (XkbBI_MinorError,0);
+    }
+}
+#endif /* TEK_MODE */
+
+
 /* ARGSUSED */
 void HandleSecure(w, event, params, param_count)
     Widget w;
@@ -1517,13 +1716,71 @@ void HandleSecure(w, event, params, param_count)
     String *params;		/* [0] = volume */
     Cardinal *param_count;	/* 0 or 1 */
 {
-    Time time = CurrentTime;
+    Time ev_time = CurrentTime;
 
     if ((event->xany.type == KeyPress) ||
 	(event->xany.type == KeyRelease))
-	time = event->xkey.time;
+	ev_time = event->xkey.time;
     else if ((event->xany.type == ButtonPress) ||
 	     (event->xany.type == ButtonRelease))
-      time = event->xbutton.time;
-    DoSecureKeyboard (time);
+        ev_time = event->xbutton.time;
+    DoSecureKeyboard (ev_time);
 }
+
+int
+conv_ksc_to_ct(str, len, des)
+Char *str;
+int len;
+Char *des;
+{
+	int hangul = 0;
+	int ctlen = 0;
+	if (len <= 0) len = strlen(str);
+	if (des) {
+		while (len > 0) {
+			if (*str & 0x80) {
+				if (!hangul) { 
+					*des++ = 27;
+					*des++ = '$';
+					*des++ = '(';
+					*des++ = 'C';
+					hangul = 1;
+				}
+				*des++ = *str++ & 0x7f;
+				*des++ = *str++ & 0x7f;
+				len -= 2;
+			} else {
+				if (hangul) {
+					*des++ = 27;
+					*des++ = '(';
+					*des++ = 'B';
+					hangul = 0;
+				}
+				*des++ = *str++;
+				len--;
+			}
+		}
+	} else {
+		while (len > 0) {
+			if (*str & 0x80) {
+				if (!hangul) { 
+					ctlen += 4;
+					hangul = 1;
+				}
+				ctlen += 2;
+				str += 2;
+				len -= 2;
+			} else {
+				if (hangul) {
+					ctlen += 3;
+					hangul = 0;
+				}
+				ctlen++;
+				str++;
+				len--;
+			}
+		}
+	}
+	return ctlen;
+}
+
